import os
from typing import Optional, Dict

import click

from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort
from montecarlodata.queries.onboarding import ADD_CONNECTION_MUTATION
from montecarlodata.queries.user import GET_USER_QUERY
from montecarlodata.utils import GqlWrapper, AwsClientWrapper


class DataLakeOnBoardService:
    _DC_S3_MECH = 'dc-s3'
    _DEFAULT_WAREHOUSE_TYPE = 'data-lake'

    def __init__(self, config: Config,
                 request_wrapper: Optional[GqlWrapper] = None,
                 aws_wrapper: Optional[AwsClientWrapper] = None):
        self._config = config
        self._request_wrapper = request_wrapper or GqlWrapper(mcd_id=config.mcd_id,
                                                              mcd_token=config.mcd_token)
        self._aws_wrapper = aws_wrapper or AwsClientWrapper(profile_name=config.aws_profile,
                                                            region_name=config.aws_region)

        self._user = self._request_wrapper.make_request(GET_USER_QUERY)  # get user info
        self._set_user_props()

    def _handle_cert(self, cert_prefix: str, options: Dict) -> None:
        """
        Handles cert payload from either an s3 path or file. Uploading the latter
        Options is updated if successful.
        """
        if options.get('cert_file') is not None:
            bucket_name = self._get_dc_property(prop='PrivateS3BucketArn').split(':')[5]  # get name from arn
            object_name = os.path.join(cert_prefix, os.path.basename(options['cert_file']))
            self._aws_wrapper.upload_file(bucket_name=bucket_name, object_name=object_name,
                                          file_path=options['cert_file'])

            click.echo(f"Uploaded '{options['cert_file']}' to s3://{bucket_name}/{object_name}")
            options['cert_s3'] = object_name

        if options.get('cert_s3') is not None:
            # reformat to generic options and specify a mechanism
            options['ssl_options'] = {'mechanism': self._DC_S3_MECH, 'cert': options.pop('cert_s3')}
        options.pop('cert_file', None)

    def _validate_connection(self, query: str, response_field: str, **kwargs) -> str:
        """
        Validate connection before adding using the expected gql response_field (e.g. testPrestoCredentials)
        """
        variables = self._request_wrapper.convert_snakes_to_camels(kwargs)
        temp_path = self._request_wrapper.make_request(
            query=query, variables=variables).get(response_field, {}).get('key')

        if temp_path is not None:
            click.echo('Connection validated!')
            return temp_path
        complain_and_abort('Connection failed!')

    def _add_connection(self, temp_path: str, connection_type: str) -> bool:
        """
        Add connection and setup any associated jobs
        """
        connection_request = {'key': temp_path, 'connectionType': connection_type}
        num_of_warehouses = len(self._warehouses)

        # If no warehouse has been created for this account specify the type, otherwise pass the existing warehouse id
        if num_of_warehouses == 0:
            connection_request['createWarehouseType'] = self._DEFAULT_WAREHOUSE_TYPE
        elif num_of_warehouses == 1:
            connection_request['dwId'] = self._warehouses[0]['uuid']
        else:
            complain_and_abort('More than one warehouse is not supported')

        response = self._request_wrapper.make_request(query=ADD_CONNECTION_MUTATION, variables=connection_request)
        connection_id = response.get('addConnection', {}).get('connection', {}).get('uuid')
        if connection_id is not None:
            click.echo(f"Success! Added connection for {connection_type.capitalize()}.")
            return True
        complain_and_abort('Failed to add connection!')

    def _set_user_props(self) -> None:
        """
        Set relevant user props like DC and warehouse to instance
        """
        self._dc_outputs = None
        self._collectors = self._user['getUser']['account'].get('dataCollectors', [{}])
        self._dc_index = self._get_active_collector()
        if self._dc_index is None:
            complain_and_abort('No active collector found')
        self._warehouses = self._user['getUser']['account']['dataCollectors'][self._dc_index].get('warehouses')

    def _get_dc_property(self, prop: str) -> Optional[str]:
        """
        Retrieve property from DC stack outputs
        """
        self._dc_outputs = self._dc_outputs or self._aws_wrapper.get_stack_outputs(
            self._collectors[self._dc_index]['stackArn'])  # cache lookup
        for output in self._dc_outputs:
            if output['OutputKey'] == prop:
                return output['OutputValue']

    def _get_active_collector(self) -> Optional[int]:
        """
        Get active collector - currently only one active collector per account is supported
        """
        for idx, collector in enumerate(self._collectors):
            if collector.get('active'):
                return idx



import json
import re
import traceback
from typing import Dict, Optional

import boto3
import click
import requests
from retry import retry

import montecarlodata.settings as settings
from montecarlodata.errors import manage_errors, echo_error


class Wrapper:
    def __init__(self, abort_on_error: Optional[bool] = True):
        self._abort_on_error = abort_on_error


class GqlWrapper(Wrapper):
    def __init__(self, mcd_id: str, mcd_token: str):
        self._mcd_id = mcd_id
        self._mcd_token = mcd_token
        self._endpoint = settings.MCD_API_ENDPOINT

        super().__init__()

    def make_request(self, query: str, variables: Optional[Dict] = None) -> Optional[Dict]:
        """
        Make a GraphQl request to the MCD API
        """
        headers = {'x-mcd-id': self._mcd_id, 'x-mcd-token': self._mcd_token, 'Content-Type': 'application/json'}
        payload = {'query': query, 'variables': variables or {}}
        try:
            response = self._make_request(headers=headers, payload=payload)
        except requests.exceptions.ConnectionError:
            echo_error(f"Failed to make request to '{self._endpoint}'", [traceback.format_exc()])
        except requests.exceptions.Timeout:
            echo_error(f"Timed out request to '{self._endpoint}'", [traceback.format_exc()])
        else:
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as err:
                echo_error(err, json.loads(response.text).get('errors'))
            else:
                return json.loads(response.text)['data']
        if self._abort_on_error:
            raise click.Abort()

    @retry(tries=3, delay=0.2, backoff=2, max_delay=1)
    def _make_request(self, headers: Dict, payload: Dict) -> requests.request:
        return requests.post(self._endpoint, json=payload, headers=headers)

    @staticmethod
    def convert_snakes_to_camels(dict_: Dict) -> Dict:
        """
        Converts dictionary keys from snake_case to camelCase as Gql is very opinionated
        """
        return {re.sub(r'_([a-z])', lambda x: x.group(1).upper(), k): v for k, v in dict_.items()}


class AwsClientWrapper(Wrapper):
    def __init__(self, profile_name: Optional[str] = None, region_name: Optional[str] = None,
                 session: Optional[boto3.session.Session] = None):
        super().__init__()

        self._session = session or self._get_session(profile_name=profile_name, region_name=region_name)
        self._test_if_region_is_set()

    @manage_errors
    def get_stack_outputs(self, stack_id: str) -> Dict:
        """
        Retrieve stack outputs from DC
        """
        return self._session.client('cloudformation').describe_stacks(StackName=stack_id)['Stacks'][0]['Outputs']

    @manage_errors
    def upload_file(self, bucket_name: str, object_name: str, file_path: str) -> None:
        """
        Upload a file to s3://bucket/object from file path
        """
        self._session.client('s3').upload_file(file_path, bucket_name, object_name)

    @manage_errors
    def _test_if_region_is_set(self) -> None:
        """
        Test if region is set by creating a client. Abort on failure
        """
        self._session.client('cloudformation')  # throws NoRegion if not set

    @manage_errors
    def _get_session(self, profile_name: str, region_name: str) -> Optional[boto3.session.Session]:
        """
        Create a session using named profile/region if set. Uses AWS defaults if not set.
        """
        session_args = (('profile_name', profile_name), ('region_name', region_name))
        return boto3.session.Session(**{k: v for k, v in session_args if v})  # throws ProfileNotFound if invalid



from montecarlodata.config import Config
from montecarlodata.onboarding.data_lake.data_lakes import DataLakeOnBoardService
from montecarlodata.queries.onboarding import TEST_PRESTO_CRED_MUTATION


class PrestoOnBoardService(DataLakeOnBoardService):
    _CERT_PREFIX = 'certificates/presto/'
    _EXPECTED_GQL_RESPONSE_FIELD = 'testPrestoCredentials'
    _CONNECTION_TYPE = 'presto'

    def __init__(self, config: Config, **kwargs):
        super().__init__(config, **kwargs)

    def on_board_presto_sql(self, **kwargs) -> None:
        """
        On-board a presto-sql connection by validating and adding a connection.
        Also, optionally uploads a certificate to the DC bucket.
        """
        self._handle_cert(cert_prefix=self._CERT_PREFIX, options=kwargs)
        temp_path = self._validate_connection(query=TEST_PRESTO_CRED_MUTATION,
                                              response_field=self._EXPECTED_GQL_RESPONSE_FIELD, **kwargs)
        self._add_connection(temp_path=temp_path, connection_type=self._CONNECTION_TYPE)



# Queries related to users

GET_USER_QUERY = """
query getUser {
  getUser {
    account {
      uuid
      dataCollectors {
        uuid
        stackArn
        active
        warehouses {
          uuid
        }
      }
    }
  }
}
"""



# Queries related to on-boarding

TEST_PRESTO_CRED_MUTATION = """
mutation testPrestoCredentials($catalog:String, $host:String, $httpScheme:String, $password:String, $port:Int, $schema:String, $sslOptions:SslInputOptions, $user:String) {
  testPrestoCredentials(catalog:$catalog, host:$host, httpScheme:$httpScheme, password:$password, port:$port, schema:$schema, sslOptions: $sslOptions, user:$user) {
    key
  }
}
"""

ADD_CONNECTION_MUTATION = """
mutation addConnection($connectionType:String!, $createWarehouseType:String, $dwId:UUID, $jobTypes:[String], $key:String!) {
  addConnection(connectionType:$connectionType, createWarehouseType:$createWarehouseType, dwId:$dwId, jobTypes:$jobTypes, key:$key){
    connection {
      uuid
    }
  }
}
"""



import click
import click_config_file

import montecarlodata.settings as settings
from montecarlodata.errors import complain_and_abort
from montecarlodata.onboarding.data_lake.presto import PrestoOnBoardService


@click.group(help='On-board Monte Carlo warehouse or data-lake.')
def on_boarding():
    """
    Group for any on-board sub-commands
    """
    pass


@on_boarding.command(help='Setup Presto SQl Connection. For health queries like metrics and distribution.')
@click.pass_obj
@click.option('--host', help='Hostname.', required=True)
@click.option('--port', help='HTTP port (default=8889).', default=8889, type=click.INT)
@click.option('--user', help='Username with access to catalog/schema.', required=False)
@click.password_option('--password', help='User\'s password.', prompt='Password for user (enter to skip)',
                       default='', required=False)
@click.option('--catalog', help='Mount point to access data source.', required=False)
@click.option('--schema', help='Schema to access.', required=False)
@click.option('--http-scheme', help='Scheme for authentication.',
              type=click.Choice(['http', 'https'], case_sensitive=True), required=True)
@click.option('--cert-file', help='Local SSL certificate file to upload to collector.', required=False,
              type=click.Path(dir_okay=False, exists=True))
@click.option('--cert-s3', help='Object path (key) to a certificate already uploaded to the collector.',
              required=False)
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def presto_sql(ctx, host, port, user, password, catalog, schema, http_scheme, cert_file, cert_s3):
    """
    On-board a presto sql connection
    """
    if not password:
        password = None  # make explicitly null if not set. Prompts can't be None
    if cert_file is not None and cert_s3 is not None:
        complain_and_abort('Can have a cert-file or cert-s3-path, but not both')
    PrestoOnBoardService(config=ctx['config']).on_board_presto_sql(host=host, port=port, user=user, password=password,
                                                                   catalog=catalog, schema=schema,
                                                                   http_scheme=http_scheme, cert_file=cert_file,
                                                                   cert_s3=cert_s3)
